/* Pseudo device library routines Copyright (C), Lawrence Foard, 1992 
 * May be used under the gnu library license
 * I may be contacted at:
 * lfoard@turing.org, entropy@world.std.com, 
 * entropy on the VWIS Linux support BBS 1-508-793-9568 
 * Snail mail:
 * Lawrence Foard
 * 217 Park Ave. 
 * Box #108
 * Worcester MA 10609
 */
#include <stdio.h>
#include <fcntl.h>
#include <sys/pseudo.h>
#include <filters.h>

/* block size translation object */

BLOCKING *open_blocking(int bsize,int (*read_fun)(),int (*write_fun)(),void *data)
 {
  BLOCKING *blocking=(BLOCKING *) malloc(sizeof(BLOCKING));
  blocking->bsize=bsize;
  blocking->tmp_blk=(char *) malloc(bsize);
  blocking->read_fun=read_fun;
  blocking->write_fun=write_fun;
  blocking->data=data;
  return(blocking);
 }

void close_blocking(BLOCKING *blocking)
 {
  free(blocking->tmp_blk);
  free(blocking);
 }
 
int read_blocking(BLOCKING *blocking,int pos,char *buf,int count)
 {
  unsigned int block;
  unsigned int offset;
  unsigned int chars;
  int read=0;
/*  printf("unblocked read %d %d\n",pos,count);*/
  /* see if we where lucky enough to get an already aligned read */
  if ((count==blocking->bsize) && !(pos % blocking->bsize))
   return(blocking->read_fun(blocking->data,pos,buf,count));
  block=pos/blocking->bsize;
  offset=pos % blocking->bsize;
  /* following code adapted from Linux kernel (written by Linus Torvalds) */
  while (count>0) 
   {
    chars=blocking->bsize-offset;
    if (chars>count)
     chars=count;
    /* if reading whole blocks read directly into buffer */
    if (chars==blocking->bsize)
     {
      if (!blocking->read_fun(blocking->data,block*blocking->bsize,buf,blocking->bsize))
       return(read);
     }
    else
     {
      if (!blocking->read_fun(blocking->data,block*blocking->bsize,blocking->tmp_blk,blocking->bsize))
       return(read);
      memcpy(buf,blocking->tmp_blk+offset,chars);
     }
    block++;
    offset=0;
    read+=chars;
    count-=chars;
    buf+=chars;
   }
  return(read);
 }

int write_blocking(BLOCKING *blocking,int pos,char *buf,int count)
 {
  int block;
  int offset;
  int chars;
  int written=0;
/*  printf("unblocked write %d %d\n",pos,count);*/
  /* see if we where lucky enough to get an already aligned write */
  if ((count==blocking->bsize) && !(pos % blocking->bsize))
   return(blocking->write_fun(blocking->data,pos,buf,count));
  /* following code adapted from Linux kernel (written by Linus Torvalds) */
  block=pos/blocking->bsize;
  offset=pos % blocking->bsize;
  while (count>0) 
   {
    chars=blocking->bsize-offset;
    if (chars>count)
     chars=count;
    if (chars!=blocking->bsize)
     /* need to read a block */
     if (!blocking->read_fun(blocking->data,block*blocking->bsize,blocking->tmp_blk,blocking->bsize))
      return(written);
/*    block++;*/
/*    p=offset+blocking->tmp_blk;*/
    written+=chars;
    count-=chars;
    /* if an exact block then write directly from input buffer */
    if (chars==blocking->bsize)
     {
      if (!blocking->write_fun(blocking->data,block*blocking->bsize,buf,blocking->bsize))
       return(written);
     }
    /* have to copy into partial buffer */
    else
     {
      memcpy(blocking->tmp_blk+offset,buf,chars);
      if (!blocking->write_fun(blocking->data,block*blocking->bsize,blocking->tmp_blk,blocking->bsize))
       return(written);
     }
    offset=0;
    buf+=chars;
    block++;
   }
  return written;
 }
